from __future__ import absolute_import, division, print_function

import pickle
import logging

import datasets


_CITATION = """\
@misc{kraljevic2020multidomain,
      title={Multi-domain Clinical Natural Language Processing with MedCAT: the Medical Concept Annotation Toolkit}, 
      author={Zeljko Kraljevic and Thomas Searle and Anthony Shek and Lukasz Roguski and Kawsar Noor and Daniel Bean and Aurelie Mascio and Leilei Zhu and Amos A Folarin and Angus Roberts and Rebecca Bendayan and Mark P Richardson and Robert Stewart and Anoop D Shah and Wai Keong Wong and Zina Ibrahim and James T Teo and Richard JB Dobson},
      year={2020},
      eprint={2010.01165},
      archivePrefix={arXiv},
      primaryClass={cs.CL}
}
"""

_DESCRIPTION = """\
Takes as input a pickled dict of annotated documents from MedCAT. The format should be:
    {'document_id': {'entities': <entities>, ...}
Where entities is the output from medcat.get_entities(<...>)['entities']
"""

class MedCATAnnotationsConfig(datasets.BuilderConfig):
    """ BuilderConfig for MedCATAnnotations.

        Args:
          **kwargs: keyword arguments forwarded to super.
    """

    def __init__(self, **kwargs):
        super(MedCATAnnotationsConfig, self).__init__(**kwargs)


class MedCATAnnotations(datasets.GeneratorBasedBuilder):
    """MedCATAnnotations: Output of MedCAT"""

    BUILDER_CONFIGS = [
        MedCATAnnotationsConfig(
            name="pickle",
            version=datasets.Version("1.0.0", ""),
            description="Pickled output from MedCAT",
        ),
    ]


    def _info(self):
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "id": datasets.Value("int32"),
                    "document_id": datasets.Value("string"),
                    "context_left": datasets.Value("string"),
                    "context_right": datasets.Value("string"),
                    "context_center": datasets.Value("string"),
                }
            ),
            # No default supervised_keys (as we have to pass both question
            # and context as input).
            supervised_keys=None,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                gen_kwargs={
                    "filepath": self.config.data_files,
                },
            ),
        ]

    def _generate_examples(self, filepath):
        """This function returns the examples in the raw (text) form."""
        logging.info("generating examples from = %s", filepath)
        with open(filepath, 'rb') as f:
            docs = pickle.load(f)
            for doc_id in docs:
                doc = docs[doc_id]
                for id, entity in doc['entities'].items():
                    yield "{}|{}".format(doc_id, entity['id']), {
                            'id': int(id),
                            'document_id': str(doc_id),
                            'context_left': "".join(entity['context_left']),
                            'context_right': "".join(entity['context_right']),
                            'context_center': "".join(entity['context_center']),
                            }



def check_filters(cui, filters):
    r''' Checks is a CUI in the filters

    Args:
        cui
        filters

    Return:
        bool:
            True if in filters else False
    '''
    if cui in filters.get('cuis', {}) or not filters.get('cuis', {}):
        if cui not in filters.get('cuis_exclude', {}):
            return True
        else:
            return False
    else:
        return False

def process_old_project_filters(cuis, type_ids, cdb):
    cui_filter = set()
    if cuis is not None and cuis:
        cui_filter.update([x.strip() for x in cuis.split(",")])
    if type_ids is not None and type_ids:
        type_ids = [x.strip().upper() for x in type_ids.split(",")]

        # Convert type_ids to cuis
        if 'type_id2cuis' in cdb.addl_info:
            for type_id in type_ids:
                if type_id in cdb.addl_info['type_id2cuis']:
                    cui_filter.update(cdb.addl_info['type_id2cuis'][type_id])
                else:
                    raise Exception("Impossible to create filters, disable them.")
        else:
            raise Exception("Impossible to create filters, disable them.")

    return cui_filter




""" I would just ignore this whole class, it's just a lot of rules that work nicely for CDB
once the software is trained the main thing are the context vectors.
"""
import numpy as np
import operator
from spacy.tokens import Span
import logging

log = logging.getLogger(__name__)

def maybe_annotate_name(name, tkns, doc, cdb, config, label='concept'):
    r''' Given a name it will check should it be annotated based on config rules. If yes
    the annotation will be added to the doc._.ents array.

    Args:
        name (`str`):
            The name found in the text of the document.
        tkns (`List[spacy.tokens.Token]`):
            Tokens that belong to this name in the spacy document.
        doc (`spacy.tokens.Doc`):
            Spacy document to be annotated with named entities.
        cdb (`medcat.cdb.CDB`):
            Concept database.
        config (`medcat.config.Config`):
            Global config for medcat.
        label (`str`):
            Label for this name (usually `concept` if we are using a vocab based approach).
    '''

    log.debug("Maybe annotating name: {}".format(name))
    if len(name) >= config.ner['min_name_len']:
        # Check the upper case limit, last part checks is it one token and uppercase
        if len(name) >= config.ner['upper_case_limit_len'] or (len(tkns) == 1 and tkns[0].is_upper):
            # Everything is fine, mark name
            entity = Span(doc, tkns[0].i, tkns[-1].i + 1, label=label)
            # Only set this property when using a vocab approach and where this name
            #fits a name in the cdb. All standard name entity recognition models will not set this.
            entity._.detected_name = name
            entity._.link_candidates = cdb.name2cuis[name]
            entity._.id = len(doc._.ents)
            entity._.confidence = -1 #  This does not calculate confidence
            # Append the entity to the document
            doc._.ents.append(entity)

            # Not necessary, but why not
            log.debug("NER detected an entity." +
                      "\n\tDetected name: {}".format(entity._.detected_name) + 
                      "\n\tLink candidates: {}\n".format(entity._.link_candidates))
                    
            return entity

    return None

"""
def check_disambiguation_status(name, cuis, config):
    if len(name) < config.linking['disamb_len_limit']:
        return True
    elif len(cuis) == 1:
        if cdb.name2cui2status[name][cuis[0]] != 'N':
            return True
        else:
            return False
    else:
        for cui in cuis:
            if self.cdb.name2cui2status[name][cui] == 'P':

"""
"""
class (object):
    def __init__(self, cdb, spacy_cat):
        self.cdb = cdb
        self._cat = spacy_cat
        self.pref_names = set(cdb.cui2pref_name.values())


    def CheckAnnotation(self, name, tkns, doc, to_disamb, doc_words):
        # Do not add annotations here, but just return what needs to be done



        # Is the detected name uppercase and only one token
        # Must be one token, multiple is very dangerous
        # First check length limit and uppercase limit
        elif len(name) < config.length_limit:
            # Disambiguate
            return 'disambiguate'
        elif self.cdb.name2status[name] == 'A': # Check the cdb 
            if len(self.cdb.name2cui[name]) == 1:
                # Links to only one CUI
                return 'annotate'
            else:
                # Links to multiple CUIs
                return 'disambiguate'
        elif self.cdb.name2status == 'N':
            return 'disambiguate'
        elif self.cdb.
            return 'annotate'


### This is for annotation, should be moved

        if is_train:
            if len(name) < config.disamb_length_limit:
                cuis = self.cdb.name2cuis[name]
                if len(cuis) == 1:
                    if self.cdb.name2cui2status[name][cuis[0]] != 'N':
                        return cuis[0]
                else:
                    for cui in cuis:
                        if self.cdb.name2cui2status[name][cui] == 'P':
                            # Means this name should be used for training as it nearly always links to
                            #the concept with this CUI
                            return cui # Break the loop, one name marked with 'P' linkes to max 1 concept
            return None

        else:
            cuis = self.cdb.name2cuis[name]
            if len(name) < config.disamb_length_limit:
                return disambiguate()
            elif len(cuis) == 1:
                if self.cdb.name2cui2status[name][cuis[0]] == 'N':
                    return disambiguate()
                else:
                    return cuis[0]
            else:
                # Returns None if below thrashold
                return disambiguate(doc, ent, cuis, cdb, config)

# Disambiguate function should check the cut-offs based on min context similarity
#- Reward P, but punish N, leave 0 for A
#- Take 1000 most frequent words, set length limit and make them all 'N'
### End of annotation


        if len(name) > 1 or one_tkn_upper:
            if name in self.cdb.name_isunique:
                # Is the number of tokens matching for short words
                if len(name) >= 7 or len(tkns) in self.cdb.name2ntkns[name]:
                    if self.cdb.name_isunique[name]:
                        # Annotate
                        cui = list(self.cdb.name2cui[name])[0]
                        self._cat._add_ann(cui, doc, tkns, acc=1, name=name)
                    else:
                        to_disamb.append((list(tkns), name))
                else:
                    # For now ignore if < 7 and tokens don't match
                    #to_disamb.append((list(tkns), name))
                    pass
            else:
                # Is the number of tokens matching for short words
                if len(name) > 7 or len(tkns) in self.cdb.name2ntkns[name]:
                    if len(self.cdb.name2cui[name]) == 1 and len(name) > 2:
                        # There is only one concept linked to this name and has
                        #more than 2 characters
                        cui = list(self.cdb.name2cui[name])[0]
                        self._cat._add_ann(cui, doc, tkns, acc=1, name=name)
                    elif self._cat.train and name in self.pref_names and len(name) > 3:
                        # If training use prefered names as ground truth
                        cuis = self.cdb.name2cui[name]
                        for cui in cuis:
                            if name == self.cdb.cui2pref_name.get(cui, 'nan-nan'):
                                self._cat._add_ann(cui, doc, tkns, acc=1, name=name)
                    else:
                        to_disamb.append((list(tkns), name))
                else:
                    # For now ignore
                    #to_disamb.append((list(tkns), name))
                    pass
"""



import pandas
import spacy
import numpy as np
import datetime
import logging
from functools import partial
import re

from medcat.pipe import Pipe
from medcat.cdb import CDB
from medcat.preprocessing.tokenizers import spacy_split_all
from medcat.preprocessing.cleaners import prepare_name
from medcat.preprocessing.taggers import tag_skip_and_punct
from medcat.utils.loggers import add_handlers

PH_REMOVE = re.compile("(\s)\([a-zA-Z]+[^\)\(]*\)($)")

class CDBMaker(object):
    r''' Given a CSV as shown in https://github.com/CogStack/MedCAT/tree/master/examples/<example> it creates a CDB or
    updates an exisitng one.

    Args:
        config (`medcat.config.Config`):
            Global config for MedCAT.
        cdb (`medcat.cdb.CDB`, optional):
            If set the `CDBMaker` will updat the existing `CDB` with new concepts in the CSV.
        name_max_words (`int`, defaults to `20`):
            Names with more words will be skipped during the build of a CDB
    '''
    log = logging.getLogger(__package__)
    log = add_handlers(log)
    def __init__(self, config, cdb=None, name_max_words=20):
        self.config = config
        # Set log level
        self.log.setLevel(self.config.general['log_level'])

        # To make life a bit easier
        self.cnf_cm = config.cdb_maker

        if cdb is None:
            self.cdb = CDB(config=self.config)
        else:
            self.cdb = cdb

        # Build the required spacy pipeline
        self.nlp = Pipe(tokenizer=spacy_split_all, config=config)
        self.nlp.add_tagger(tagger=partial(tag_skip_and_punct, config=self.config),
                            name='skip_and_punct',
                            additional_fields=['is_punct'])


    def prepare_csvs(self, csv_paths, sep=',', encoding=None, escapechar=None, index_col=False, full_build=False, only_existing_cuis=False, **kwargs):
        r''' Compile one or multipe CSVs into a CDB.

        Args:
            csv_paths (`List[str]`):
                An array of paths to the csv files that should be processed
            full_build (`bool`, defautls to `True`):
                If False only the core portions of the CDB will be built (the ones required for
                the functioning of MedCAT). If True, everything will be added to the CDB - this
                usually includes concept descriptions, various forms of names etc (take care that
                this option produces a much larger CDB).
            sep (`str`, defaults to `,`):
                If necessarya a custom separator for the csv files
            encoding (`str`, optional):
                Encoing to be used for reading the CSV file
            escapechar (`str`, optional):
                Escapechar for the CSV
            index_col (`bool`, defaults_to `False`):
                Index column for pandas read_csv
            only_existing_cuis (`bool`, defaults to False):
                If True no new CUIs will be added, but only linked names will be extended. Mainly used when
                enriching names of a CDB (e.g. SNOMED with UMLS terms).
        Return:
            `medcat.cdb.CDB` with the new concepts added.

        Note:
            **kwargs:
                Will be passed to pandas for CSV reading
            csv:
                Examples of the CSV used to make the CDB can be found on [GitHub](link)
        '''

        useful_columns = ['cui', 'name', 'ontologies', 'name_status', 'type_ids', 'description']
        name_status_options = {'A', 'P', 'N'}

        for csv_path in csv_paths:
            # Read CSV, everything is converted to strings
            df = pandas.read_csv(csv_path, sep=sep, encoding=encoding, escapechar=escapechar, index_col=index_col, dtype=str, **kwargs)
            df = df.fillna('')

            # Find which columns to use from the CSV
            cols = []
            col2ind = {}
            for col in list(df.columns):
                if str(col).lower().strip() in useful_columns:
                    col2ind[str(col).lower().strip()] = len(cols)
                    cols.append(col)

            self.log.info("Started importing concepts from: {}".format(csv_path))
            _time = None # Used to check speed
            _logging_freq = np.ceil(len(df[cols]) / 100)
            for row_id, row in enumerate(df[cols].values):
                if row_id % _logging_freq == 0:
                    # Print some stats
                    if _time is None:
                        # Add last time if it does not exist
                        _time = datetime.datetime.now()
                    # Get current time
                    ctime = datetime.datetime.now()
                    # Get time difference
                    timediff = ctime - _time
                    self.log.info("Current progress: {:.0f}% at {:.3f}s per {} rows".format(
                        (row_id / len(df)) * 100, timediff.microseconds/10**6 + timediff.seconds, (len(df[cols]) // 100)))
                    # Set previous time to current time
                    _time = ctime

                # This must exist
                cui = row[col2ind['cui']].strip().upper()

                if not only_existing_cuis or (only_existing_cuis and cui in self.cdb.cui2names):
                    if 'ontologies' in col2ind:
                        ontologies = set([ontology.strip() for ontology in row[col2ind['ontologies']].upper().split(self.cnf_cm['multi_separator']) if
                                         len(ontology.strip()) > 0])
                    else:
                        ontologies = set()

                    if 'name_status' in col2ind:
                        name_status = row[col2ind['name_status']].strip().upper()

                        # Must be allowed
                        if name_status not in name_status_options:
                            name_status = 'A'
                    else:
                        # Defaults to A - meaning automatic
                        name_status = 'A'

                    if 'type_ids' in col2ind:
                        type_ids = set([type_id.strip() for type_id in row[col2ind['type_ids']].upper().split(self.cnf_cm['multi_separator']) if
                                        len(type_id.strip()) > 0])
                    else:
                        type_ids = set()

                    # Get the ones that do not need any changing
                    if 'description' in col2ind:
                        description = row[col2ind['description']].strip()
                    else:
                        description = ""

                    # We can have multiple versions of a name
                    names = {} # {'name': {'tokens': [<str>], 'snames': [<str>]}}

                    raw_names = [raw_name.strip() for raw_name in row[col2ind['name']].split(self.cnf_cm['multi_separator']) if 
                                 len(raw_name.strip()) > 0]
                    for raw_name in raw_names:
                        raw_name = raw_name.strip()
                        prepare_name(raw_name, self.nlp, names, self.config)

                        if self.config.cdb_maker.get('remove_parenthesis', 0) > 0 and name_status == 'P':
                            # Should we remove the content in parenthesis from primary names and add them also
                            raw_name = PH_REMOVE.sub(" ", raw_name).strip()
                            if len(raw_name) >= self.config.cdb_maker['remove_parenthesis']:
                                prepare_name(raw_name, self.nlp, names, self.config)

                    self.cdb.add_concept(cui=cui, names=names, ontologies=ontologies, name_status=name_status, type_ids=type_ids,
                                         description=description, full_build=full_build)
                    # DEBUG
                    self.log.debug("\n\n**** Added\n CUI: {}\n Names: {}\n Ontologies: {}\n Name status: {}\n".format(cui, names, ontologies, name_status) + \
                                   " Type IDs: {}\n Description: {}\n Is full build: {}".format(
                                   type_ids, description, full_build))

        return self.cdb



import numpy as np
import pickle

class Vocab(object):
    r''' Vocabulary used to store word embeddings for context similarity
    calculation. Also used by the spell checker - but not for fixing the spelling
    only for checking is something correct.

    Properties:
        vocab (dict):
            Map from word to attributes, e.g. {'house': {'vec': <np.array>, 'cnt': <int>, ...}, ...}
        index2word (dict):
            From word to an index - used for negative sampling
        vec_index2word (dict):
            Same as index2word but only words that have vectors
        unigram_table (dict):
            Negative sampling.
    '''
    def __init__(self):
        self.vocab = {}
        self.index2word = {}
        self.vec_index2word = {}
        self.unigram_table = []


    def inc_or_add(self, word, cnt=1, vec=None):
        r''' Add a word or incrase its count.

        Args:
            word (str):
                Word to be added
            cnt (str):
                By how much should the count be increased, or to wha
                should it be set if a new word.
            vec (np.array):
                Word vector
        '''
        if word not in self.vocab:
            self.add_word(word, cnt, vec)
        else:
            self.inc_wc(word, cnt)


    def remove_all_vectors(self):
        r''' Remove all stored vector representations
        '''
        self.vec_index2word = {}

        for word in self.vocab:
            self.vocab[word]['vec'] = None


    def remove_words_below_cnt(self, cnt):
        r''' Remove all words with frequency below cnt.

        Args:
            cnt (int):
                Word count limit.
        '''
        print("Words before removal: " + str(len(self.vocab)))
        for word in list(self.vocab.keys()):
            if self.vocab[word]['cnt'] < cnt:
                del self.vocab[word]
        print("Words after removal : " + str(len(self.vocab)))

        # Rebuild index2word and vec_index2word
        self.index2word = {}
        self.vec_index2word = {}
        for word in self.vocab.keys():
            ind = len(self.index2word)
            self.index2word[ind] = word
            self.vocab[word]['ind'] = ind

            if self.vocab[word]['vec'] is not None:
                self.vec_index2word[ind] = word


    def inc_wc(self, word, cnt=1):
        r''' Incraese word count by cnt.

        Args:
            word (str):
                For which word to increase the count
            cnt:
                By how muhc to incrase the count
        '''
        self.item(word)['cnt'] += cnt


    def add_vec(self, word, vec):
        r''' Add vector to a word.

        Args:
            word (str):
                To which word to add the vector.
            vec (np.array):
                The vector to add.
        '''
        self.vocab[word]['vec'] = vec

        ind = self.vocab[word]['ind']
        if ind not in self.vec_index2word:
            self.vec_index2word[ind] = word


    def reset_counts(self, cnt=1):
        r''' Reset the count for all word to cnt.

        Args:
            cnt (int):
                New count for all words in the vocab.
        '''
        for word in self.vocab.keys():
            self.vocab[word]['cnt'] = cnt

    def update_counts(self, tokens):
        r''' Given a list of tokens update counts for words in the vocab.

        Args:
            tokens (List[str]):
                Usually a large block of text split into tokens/words.
        '''
        for token in tokens:
            if token in self:
                self.inc_wc(token, 1)


    def add_word(self, word, cnt=1, vec=None, replace=True):
        """Add a word to the vocabulary

        Args:
            word (str):
                the word to be added, it should be lemmatized and lowercased
            cnt (int):
                count of this word in your dataset
            vec (np.array):
                the vector repesentation of the word
            replace (bool):
                will replace old vector representation
        """
        if word not in self.vocab:
            ind = len(self.index2word)
            self.index2word[ind] = word
            item = {'vec': vec, 'cnt': cnt, 'ind': ind}
            self.vocab[word] = item

            if vec is not None:
                self.vec_index2word[ind] = word
        elif replace and vec is not None:
            self.vocab[word]['vec'] = vec
            self.vocab[word]['cnt'] = cnt

            # If this word didn't have a vector before
            ind = self.vocab[word]['ind']
            if ind not in self.vec_index2word:
                self.vec_index2word[ind] = word


    def add_words(self, path, replace=True):
        """Adds words to the vocab from a file, the file
        is required to have the following format (vec being optional):
            <word>\t<cnt>[\t<vec_space_separated>]

        e.g. one line: the word house with 3 dimensional vectors
            house   34444   0.3232 0.123213 1.231231

        Args:
            path (str):
                path to the file with words and vectors
            replace (bool):
                existing words in the vocabulary will be replaced
        """
        f = open(path)

        for line in f:
            parts = line.split("\t")
            word = parts[0]
            cnt = int(parts[1].strip())
            vec = None
            if len(parts) == 3:
                vec = np.array([float(x) for x in parts[2].strip().split(" ")])

            self.add_word(word, cnt, vec)


    def make_unigram_table(self, table_size=100000000):
        r''' Make unigram table for negative sampling, look at the paper if interested
        in details.
        '''
        freqs = []
        self.unigram_table = []

        words = list(self.vec_index2word.values())
        for word in words:
            freqs.append(self[word])

        freqs = np.array(freqs)
        freqs = np.power(freqs, 3/4)
        sm = np.sum(freqs)

        for ind in self.vec_index2word.keys():
            word = self.vec_index2word[ind]
            f_ind = words.index(word)
            p = freqs[f_ind] / sm
            self.unigram_table.extend([ind] * int(p * table_size))

        self.unigram_table = np.array(self.unigram_table)


    def get_negative_samples(self, n=6, ignore_punct_and_num=False):
        r''' Get N negative samples.

        Args:
            n (int):
                How many words to return
            ignore_punct_and_num (bool):
                When returing words shold we skip punctuation and numbers.
        Returns:
            inds (List[int]):
                Indices for words in this vocabulary.
        '''
        if len(self.unigram_table) == 0:
            raise Exception("No unigram table present, please run the function vocab.make_unigram_table() first.")
        inds = np.random.randint(0, len(self.unigram_table), n)
        inds = self.unigram_table[inds]

        if ignore_punct_and_num:
            # Do not return anything that does not have letters in it
            inds = [ind for ind in inds if self.index2word[ind].upper().isupper()]

        return inds


    def __getitem__(self, word):
        return self.vocab[word]['cnt']


    def vec(self, word):
        return self.vocab[word]['vec']


    def item(self, word):
        return self.vocab[word]


    def __contains__(self, word):
        if word in self.vocab:
            return True

        return False


    def save(self, path):
        with open(path, 'wb') as f:
            pickle.dump(self.__dict__, f)


    @classmethod
    def load(cls, path):
        with open(path, 'rb') as f:
            vocab = cls()
            vocab.__dict__ = pickle.load(f)
        return vocab



""" I would just ignore this whole class, it's just a lot of rules that work nicely for CDB
once the software is trained the main thing are the context vectors.
"""
import numpy as np
import operator

class CatAnn(object):
    def __init__(self, cdb, spacy_cat):
        self.cdb = cdb
        self._cat = spacy_cat
        self.pref_names = set(cdb.cui2pref_name.values())


    def add_ann(self, name, tkns, doc, to_disamb, doc_words):
        one_tkn_upper = False
        if len(tkns) == 1 and tkns[0].is_upper:
            one_tkn_upper = True

        if len(name) > 1 or one_tkn_upper:
            if name in self.cdb.name_isunique:
                # Is the number of tokens matching for short words
                if len(name) >= 7 or len(tkns) in self.cdb.name2ntkns[name]:
                    if self.cdb.name_isunique[name]:
                        # Annotate
                        cui = list(self.cdb.name2cui[name])[0]
                        self._cat._add_ann(cui, doc, tkns, acc=1, name=name)
                    else:
                        to_disamb.append((list(tkns), name))
                else:
                    # For now ignore if < 7 and tokens don't match
                    #to_disamb.append((list(tkns), name))
                    pass
            else:
                # Is the number of tokens matching for short words
                if len(name) > 7 or len(tkns) in self.cdb.name2ntkns[name]:
                    if len(self.cdb.name2cui[name]) == 1 and len(name) > 2:
                        # There is only one concept linked to this name and has
                        #more than 2 characters
                        cui = list(self.cdb.name2cui[name])[0]
                        self._cat._add_ann(cui, doc, tkns, acc=1, name=name)
                    elif self._cat.train and name in self.pref_names and len(name) > 3:
                        # If training use prefered names as ground truth
                        cuis = self.cdb.name2cui[name]
                        for cui in cuis:
                            if name == self.cdb.cui2pref_name.get(cui, 'nan-nan'):
                                self._cat._add_ann(cui, doc, tkns, acc=1, name=name)
                    else:
                        to_disamb.append((list(tkns), name))
                else:
                    # For now ignore
                    #to_disamb.append((list(tkns), name))
                    pass



""" Representation class for CDB data
"""

import dill
import logging
import numpy as np
from typing import Dict, List, Set

from medcat.utils.matutils import unitvec, sigmoid
from medcat.utils.ml_utils import get_lr_linking
from medcat.config import Config


class CDB(object):
    """ Concept DataBase - holds all information necessary for NER+L.

    Properties:
        name2cuis (`Dict[str, List[str]]`):
            Map fro concept name to CUIs - one name can map to multiple CUIs.
        name2cuis2status (`Dict[str, Dict[str, str]]`):
            What is the status for a given name and cui pair - each name can be:
                P - Preferred, A - Automatic (e.g. let medcat decide), N - Not common.
        snames (`Set[str]`):
            All possible subnames for all concepts
        cui2names (`Dict[str, Set[str]]`):
            From cui to all names assigned to it. Mainly used for subsetting (maybe even only).
        cui2snames (`Dict[str, Set[str]]`):
            From cui to all sub-names assigned to it. Only used for subsetting.
        cui2context_vectors (`Dict[str, Dict[str, np.array]]`):
            From cui to a dictionary of different kinds of context vectors. Normally you would have here
            a short and a long context vector - they are calculated separately.
        cui2count_train (`Dict[str, int]`):
            From CUI to the number of training examples seen.
        cui2tags (`Dict[str, List[str]]`):
            From CUI to a list of tags. This can be used to tag concepts for grouping of whatever.
        cui2type_ids (`Dict[str, Set[str]]`):
            From CUI to type id (e.g. TUI in UMLS).
        cui2preferred_name (`Dict[str, str]`):
            From CUI to the preferred name for this concept.
        cui2average_confidence(`Dict[str, str]`):
            Used for dynamic thresholding. Holds the average confidence for this CUI given the training examples.
        name2count_train(`Dict[str, str]`):
            Counts how often did a name appear during training.
        addl_info (`Dict[str, Dict[]]`):
            Any additional maps that are not part of the core CDB. These are usually not needed
            for the base NER+L use-case, but can be useufl for Debugging or some special stuff.
        vocab (`Dict[str, int]`):
            Stores all the words tha appear in this CDB and the count for each one.
    """
    log = logging.getLogger(__name__)
    def __init__(self, config):
        self.config = config
        self.name2cuis = {}
        self.name2cuis2status = {}

        self.snames = set()

        self.cui2names = {}
        self.cui2snames = {}
        self.cui2context_vectors = {}
        self.cui2count_train = {}
        self.cui2tags = {} # Used to add custom tags to CUIs
        self.cui2type_ids = {}
        self.cui2preferred_name = {}
        self.cui2average_confidence = {}
        self.name2count_train = {}

        self.addl_info = {
                'cui2icd10': {},
                'cui2opcs4': {},
                'cui2ontologies': {},
                'cui2original_names': {},
                'cui2description': {},
                'type_id2name': {},
                'type_id2cuis': {},
                'cui2group': {},
                # Can be extended with whatever is necessary
                }
        self.vocab = {} # Vocabulary of all words ever in our cdb
        self._optim_params = None


    def update_cui2_average_confidence(self, cui, new_sim):
        self.cui2average_confidence[cui] = (self.cui2average_confidence[cui] * self.cui2count_train[cui] + new_sim)  / (self.cui2count_train[cui] + 1)

    def remove_names(self, cui: str, names: Dict):
        r''' Remove names from an existing concept - efect is this name will never again be used to link to this concept.
        This will only remove the name from the linker (namely name2cuis and name2cuis2status), the name will still be present everywhere else.
        Why? Because it is bothersome to remove it from everywhere, but
        could also be useful to keep the removed names in e.g. cui2names.

        Args:
            cui (`str`):
                Concept ID or unique identifer in this database.
            names (`Dict[str, Dict]`):
                Names to be removed, should look like: `{'name': {'tokens': tokens, 'snames': snames, 'raw_name': raw_name}, ...}`
        '''
        for name in names.keys():
            if name in self.name2cuis:
                if cui in self.name2cuis[name]:
                    self.name2cuis[name].remove(cui)
                if len(self.name2cuis[name]) == 0:
                    del self.name2cuis[name]

            # Remove from name2cuis2status
            if name in self.name2cuis2status:
                if cui in self.name2cuis2status[name]:
                    _ = self.name2cuis2status[name].pop(cui)
                if len(self.name2cuis2status[name]) == 0:
                    del self.name2cuis2status[name]

            # Set to disamb always if name2cuis2status is now only one CUI
            if name in self.name2cuis2status:
                if len(self.name2cuis2status[name]) == 1:
                    for _cui in self.name2cuis2status[name]:
                        if self.name2cuis2status[name][_cui] == 'A':
                            self.name2cuis2status[name][_cui] = 'N'
                        elif self.name2cuis2status[name][_cui] == 'P':
                            self.name2cuis2status[name][_cui] = 'PD'


    def add_names(self, cui: str, names: Dict, name_status: str='A', full_build: bool=False):
        r''' Adds a name to an existing concept.

        Args:
            cui (`str`):
                Concept ID or unique identifer in this database, all concepts that have
                the same CUI will be merged internally.
            names (`Dict[str, Dict]`):
                Names for this concept, or the value that if found in free text can be linked to this concept.
                Names is an dict like: `{name: {'tokens': tokens, 'snames': snames, 'raw_name': raw_name}, ...}`
            name_status (`str`):
                One of `P`, `N`, `A`
            full_build (`bool`, defaults to `False`):
                If True the dictionary self.addl_info will also be populated, contains a lot of extra information
                about concepts, but can be very memory consuming. This is not necessary for normal functioning of MedCAT.
        '''
        name_status = name_status.upper()
        if name_status not in ['P', 'A', 'N']:
            # Name status must be one of the three
            name_status = 'A'

        self.add_concept(cui=cui, names=names, ontologies=set(), name_status=name_status, type_ids=set(), description='', full_build=full_build)


    def add_concept(self, cui: str, names: Dict, ontologies: set(), name_status: str, type_ids: Set[str], description: str, full_build: bool=False):
        r'''
        Add a concept to internal Concept Database (CDB). Depending on what you are providing
        this will add a large number of properties for each concept.

        Args:
            cui (`str`):
                Concept ID or unique identifer in this database, all concepts that have
                the same CUI will be merged internally.
            names (`Dict[str, Dict]`):
                Names for this concept, or the value that if found in free text can be linked to this concept.
                Names is an dict like: `{name: {'tokens': tokens, 'snames': snames, 'raw_name': raw_name}, ...}`
            ontologies(`Set[str]`):
                ontologies in which the concept exists (e.g. SNOMEDCT, HPO)
            name_status (`str`):
                One of `P`, `N`, `A`
            type_ids (`Set[str]`):
                Semantic type identifier (have a look at TUIs in UMLS or SNOMED-CT)
            description (`str`):
                Description of this concept.
            full_build (`bool`, defaults to `False`):
                If True the dictionary self.addl_info will also be populated, contains a lot of extra information
                about concepts, but can be very memory consuming. This is not necessary for normal functioning of MedCAT.
        '''
        # Add CUI to the required dictionaries
        if cui not in self.cui2names:
            # Create placeholders 
            self.cui2names[cui] = set()
            self.cui2snames[cui] = set()
            self.cui2context_vectors[cui] = {}
            self.cui2count_train[cui] = 0
            self.cui2average_confidence[cui] = 0
            self.cui2tags[cui] = []

            # Add type_ids
            self.cui2type_ids[cui] = type_ids
        else:
            # If the CUI is already in update the type_ids
            self.cui2type_ids[cui].update(type_ids)

        # Add names to the required dictionaries
        name_info = None
        for name in names:
            name_info = names[name]
            # Extend snames
            self.snames.update(name_info['snames'])

            # Add name to cui2names
            self.cui2names[cui].add(name)
            # Extend cui2snames
            self.cui2snames[cui].update(name_info['snames'])
            # Add name to counts
            self.name2count_train[name] = 0

            if name in self.name2cuis:
                # Means we have alrady seen this name
                if cui not in self.name2cuis[name]:
                    # If CUI is not already linked do it
                    self.name2cuis[name].append(cui)

                    # At the same time it means the cui is also missing from name2cuis2status, but the 
                    #name is there
                    self.name2cuis2status[name][cui] = name_status
                elif name_status == 'P':
                    # If name_status is P overwrite whatever was the old status
                    self.name2cuis2status[name][cui] = name_status
            else:
                # Means we never saw this name
                self.name2cuis[name] = [cui]

               # Add name2cuis2status
                self.name2cuis2status[name] = {cui: name_status}


            # Add tokens to vocab
            for token in name_info['tokens']:
                if token in self.vocab:
                    self.vocab[token] += 1
                else:
                    self.vocab[token] = 1

        # Check is this a preferred name for the concept, this takes the name_info
        #dict which must have a value (but still have to check it, just in case).
        if name_info is not None:
            if name_status == 'P' and cui not in self.cui2preferred_name:
                # Do not overwrite old preferred names
                self.cui2preferred_name[cui] = name_info['raw_name']

        # Add other fields if full_build
        if full_build:
            # Use original_names as the base check because they must be added
            if cui not in self.addl_info['cui2original_names']:
                if ontologies: self.addl_info['cui2ontologies'][cui] = ontologies
                if description: self.addl_info['cui2description'][cui] = description
                self.addl_info['cui2original_names'][cui] = set([v['raw_name'] for k,v in names.items()])
            else:
                # Update existing ones
                if ontologies: self.addl_info['cui2ontologies'][cui].update(ontologies)
                if description: self.addl_info['cui2description'][cui] = description
                self.addl_info['cui2original_names'][cui].update([v['raw_name'] for k,v in names.items()])

            for type_id in type_ids:
                # Add type_id2cuis link
                if type_id in self.addl_info['type_id2cuis']:
                    self.addl_info['type_id2cuis'][type_id].add(cui)
                else:
                    self.addl_info['type_id2cuis'][type_id] = {cui}


    def add_addl_info(self, name, data, reset_existing=False):
        r''' Add data to the addl_info dictionary. This is done in a function to
        not directly access the addl_info dictionary.

        Args:
            name (`str`):
                What key should be used in the `addl_info` dictionary.
            data (`Dict[<whatever>]`):
                What will be added as the value for the key `name`
            reset_existing (`bool`):
                Should old data be removed if it exists
        '''
        if reset_existing:
            self.addl_info[name] = {}

        self.addl_info[name].update(data)


    def update_context_vector(self, cui, vectors, negative=False, lr=None, cui_count=0):
        r''' Add the vector representation of a context for this CUI.

        cui (`str`):
            The concept in question.
        vectors (`Dict[str, np.array]`):
            Vector represenation of the context, must have the format: {'context_type': np.array(<vector>), ...}
            context_type - is usually one of: ['long', 'medium', 'short']
        negative (`bool`, defaults to `False`):
            Is this negative context of positive.
        lr (`int`, optional):
            If set it will override the base value from the config file.
        cui_count (`int`, defaults to 0):
            The learning rate will be calculated based on the count for the provided CUI + cui_count.
        '''
        similarity = None
        if cui in self.cui2context_vectors:
            for context_type, vector in vectors.items():
                # Get the right context
                if context_type in self.cui2context_vectors[cui]:
                    cv = self.cui2context_vectors[cui][context_type]
                    similarity = np.dot(unitvec(cv), unitvec(vector))

                    # Get the learning rate if None
                    if lr is None:
                        lr = get_lr_linking(self.config, self.cui2count_train[cui] + cui_count, self._optim_params, similarity)

                    if negative:
                        # Add negative context
                        b = max(0, similarity) * lr
                        self.cui2context_vectors[cui][context_type] = cv*(1-b) - vector*b
                    else:
                        b = (1 - max(0, similarity)) * lr
                        self.cui2context_vectors[cui][context_type] = cv*(1-b) + vector*b

                    # DEBUG
                    self.log.debug("Updated vector embedding.\n" + \
                            "CUI: {}, Context Type: {}, Similarity: {:.2f}, Is Negative: {}, LR: {:.5f}, b: {:.3f}".format(cui, context_type,
                                similarity, negative, lr, b))
                    cv = self.cui2context_vectors[cui][context_type]
                    similarity_after = np.dot(unitvec(cv), unitvec(vector))
                    self.log.debug("Similarity before vs after: {:.5f} vs {:.5f}".format(similarity, similarity_after))
                else:
                    if negative:
                        self.cui2context_vectors[cui][context_type] = -1 * vector
                    else:
                        self.cui2context_vectors[cui][context_type] = vector

                    # DEBUG
                    self.log.debug("Added new context type with vectors.\n" + \
                            "CUI: {}, Context Type: {}, Is Negative: {}".format(cui, context_type, negative))

        if not negative:
            # Increase counter only for positive examples
            self.cui2count_train[cui] += 1


    def save(self, path):
        r''' Saves model to file (in fact it saves vairables of this class). 

        Args:
            path (`str`):
                Path to a file where the model will be saved
        '''
        with open(path, 'wb') as f:
            # No idea how to this correctly
            to_save = {}
            to_save['config'] = self.config.__dict__
            to_save['cdb'] = {k:v for k,v in self.__dict__.items() if k != 'config'}
            dill.dump(to_save, f)


    @classmethod
    def load(cls, path, config=None):
        r''' Load and return a CDB. This allows partial loads in probably not the right way at all.

        Args:
            path (`str`):
                Path to a `cdb.dat` from which to load data.
        '''
        with open(path, 'rb') as f:
            # Again no idea
            data = dill.load(f)
            if config is None:
                config = Config.from_dict(data['config'])
            # Create an instance of the CDB (empty)
            cdb = cls(config=config)

            # Load data into the new cdb instance
            for k in cdb.__dict__:
                if k in data['cdb']:
                    cdb.__dict__[k] = data['cdb'][k]

        return cdb


    def import_old_cdb_vectors(self, cdb):
        # Import context vectors
        for cui in self.cui2context_vectors:
            if cui in cdb.cui2context_vec:
                self.cui2context_vectors[cui] = {'medium': cdb.cui2context_vec[cui],
                                                 'long': cdb.cui2context_vec[cui],
                                                 'xlong': cdb.cui2context_vec[cui]}

                if cui in cdb.cui2context_vec_short:
                    self.cui2context_vectors[cui]['short'] = cdb.cui2context_vec_short[cui]

                self.cui2count_train[cui] = cdb.cui_count[cui]


    def import_old_cdb(self, cdb, import_vectors=True):
        r''' Import all data except for cuis and names from an old CDB.
        '''
        
        # Import vectors
        if import_vectors:
            self.import_old_cdb_vectors(cdb)

        # Import TUIs
        for cui in cdb.cui2names:
            self.cui2type_ids[cui] = {cdb.cui2tui.get(cui, 'unk')}

        # Import TUI to CUIs
        self.addl_info['type_id2cuis'] = cdb.tui2cuis

        # Import type_id to name
        self.addl_info['type_id2name'] = cdb.tui2name

        # Import description
        self.addl_info['cui2description'] = cdb.cui2desc

        # Import ICD10 and SNOMED
        self.addl_info['cui2snomed'] = {}
        for cui in self.cui2names:
            if cui in cdb.cui2info and 'icd10' in cdb.cui2info[cui]:
                self.addl_info['cui2icd10'][cui] = cdb.cui2info[cui]['icd10']
            if cui in cdb.cui2info and 'snomed' in cdb.cui2info[cui]:
                self.addl_info['cui2snomed'][cui] = cdb.cui2info[cui]['snomed']
            if cui in cdb.cui2info and 'opcs4' in cdb.cui2info[cui]:
                self.addl_info['cui2opcs4'][cui] = cdb.cui2info[cui]['opcs4']


        # Import cui 2 ontologies
        self.addl_info['cui2ontologies'] = cdb.cui2ontos


    def import_training(self, cdb, overwrite=True):
        r''' This will import vector embeddings from another CDB. No new concepts will be added.
        IMPORTANT it will not import name maps (cui2names, name2cuis or anything else) only vectors.

        Args:
            cdb (`medcat.cdb.CDB`):
                Concept database from which to import training vectors
            overwrite (`bool`, defaults to `True`):
                If True all training data in the existing CDB will be overwritten, else
                the average between the two training vectors will be taken.

        Examples:
            >>> new_cdb.import_traininig(cdb=old_cdb, owerwrite=True)
        '''
        # Import vectors and counts
        for cui in cdb.cui2context_vectors:
            if cui in self.cui2context_vectors:
                for context_type, vector in cdb.cui2context_vectors[cui].items():
                    if overwrite or context_type not in self.cdb.cui2context_vectors[cui]:
                        self.cui2context_vectors[cui][context_type] = vector
                    else:
                        self.cui2context_vectors[cui][context_type] = (vector + self.cui2context_vectors[cui][context_type]) / 2

                # Increase the vector count
                self.cui2count_train[cui] += cdb.cui2count_train[cui]


    def reset_cui_count(self, n=10):
        r''' Reset the CUI count for all concepts that received training, used when starting new unsupervised training
        or for suppervised with annealing.

        Args:
            n (`int`, optional, defaults to 10):
                This will be set as the CUI count for all cuis in this CDB.

        Examples:
            >>> cdb.reset_cui_count()
        '''
        for cui in self.cui2count_train.keys():
            self.cui2count_train[cui] = n


    def reset_training(self):
        r''' Will remove all training efforts - in other words all embeddings that are learnt
        for concepts in the current CDB. Please note that this does not remove synonyms (names) that were
        potentially added during supervised/online learning.
        '''
        self.reset_cui_count(n=0)
        for context_type in self.cui2context_vectors:
            self.cui2context_vectors[context_type] = {}
        self.reset_concept_similarity()


    def filter_by_cui(self, cuis_to_keep):
        ''' Subset the core CDB fields (dictionaries/maps). Note that this will potenitally keep a bit more CUIs
        then in cuis_to_keep. It will first find all names that link to the cuis_to_keep and then
        find all CUIs that link to those names and keep all of them.
        This also will not remove any data from cdb.addl_info - as this field can contain data of
        unknown structure.

        Args:
            cuis_to_keep (`List[str]`):
                CUIs that will be kept, the rest will be removed (not completely, look above).
        '''
        # First get all names/snames that should be kept based on this CUIs
        names_to_keep = set()
        snames_to_keep = set()
        for cui in cuis_to_keep:
            names_to_keep.update(self.cui2names[cui])
            snames_to_keep.update(self.cui2snames[cui])

        # Based on the names get also the indirect CUIs that have to be kept
        all_cuis_to_keep = set()
        for name in names_to_keep:
            all_cuis_to_keep.update(self.name2cuis[name])

        new_name2cuis = {}
        new_name2cuis2status = {}
        new_cui2names = {}
        new_cui2snames = {}
        new_cui2context_vectors = {}
        new_cui2count_train = {}
        new_cui2tags = {} # Used to add custom tags to CUIs
        new_cui2type_ids = {}
        new_cui2preferred_name = {}

        # Subset cui2<whatever>
        for cui in all_cuis_to_keep:
            new_cui2names[cui] = self.cui2names[cui]
            new_cui2snames[cui] = self.cui2snames[cui]
            new_cui2context_vectors[cui] = self.cui2context_vectors[cui]
            new_cui2count_train[cui] = self.cui2count_train[cui]
            new_cui2tags[cui] = self.cui2tags[cui]
            new_cui2type_ids[cui] = self.cui2type_ids[cui]
            new_cui2preferred_name[cui] = self.cui2preferred_name[cui]

        # Subset name2<whatever>
        for name in names_to_keep:
            new_name2cuis[name] = self.name2cuis[name]
            new_name2cuis2status[name] = self.name2cuis2status[name]

        # Replace everything
        self.name2cuis = new_name2cuis
        self.snames = snames_to_keep
        self.name2cuis2status = new_name2cuis2status
        self.cui2names = new_cui2names
        self.cui2snames = new_cui2snames
        self.cui2context_vectors = new_cui2context_vectors
        self.cui2count_train = new_cui2count_train
        self.cui2tags = new_cui2tags
        self.cui2type_ids = new_cui2type_ids
        self.cui2preferred_name = new_cui2preferred_name


    def print_stats(self):
        r'''Print basic statistics for the CDB.
        '''
        self.log.info("Number of concepts: {:,}".format(len(self.cui2names)))
        self.log.info("Number of names:    {:,}".format(len(self.name2cuis)))
        self.log.info("Number of concepts that received training: {:,}".format(len([cui for cui in self.cui2count_train if self.cui2count_train[cui] > 0])))
        self.log.info("Number of seen training examples in total: {:,}".format(sum(self.cui2count_train.values())))
        self.log.info("Average training examples per concept:     {:.1f}".format(np.average(
            [self.cui2count_train[cui] for cui in self.cui2count_train if self.cui2count_train[cui] > 0])))


    def reset_concept_similarity(self):
        r''' Reset concept similarity matrix.
        '''
        self.addl_info['similarity'] = {}


    def most_similar(self, cui, context_type, type_id_filter=[], min_cnt=0, topn=50, force_build=False):
        r''' Given a concept it will calculate what other concepts in this CDB have the most similar
        embedding.

        Args:
            cui (`str`):
                The concept ID for the base concept for which you want to get the most similar concepts.
            context_type (`str`):
                On what vector type from the cui2context_vectors map will the similarity be calculated.
            type_id_filter (`List[str]`):
                A list of type_ids that will be used to filterout the returned results. Using this it is possible
                to limit the similarity calculation to only disorders/symptoms/drugs/...
            min_cnt (`int`):
                Minimum training examples (unsupervised+supervised) that a concept must have to be considered
                for the similarity calculation.
            topn (`int`):
                How many results to return
            force_build (`bool`, defaults to `False`):
                Do not use cached sim matrix

        Return:
            results (Dict):
                A dictionary with topn results like: {<cui>: {'name': <name>, 'sim': <similarity>, 'type_name': <type_name>,
                                                              'type_id': <type_id>, 'cnt': <number of training examples the concept has seen>}, ...}

        '''

        if 'similarity' in self.addl_info:
            if context_type not in self.addl_info['similarity']:
                self.addl_info['similarity'][context_type] = {}
        else:
            self.addl_info['similarity'] = {context_type: {}}

        sim_data = self.addl_info['similarity'][context_type]

        # Create the matrix if necessary
        if 'sim_vectors' not in sim_data or force_build:
            self.log.info("Building similarity matrix")

            sim_vectors = []
            sim_vectors_counts = []
            sim_vectors_type_ids = []
            sim_vectors_cuis = []
            for _cui in self.cui2context_vectors:
                if context_type in self.cui2context_vectors[_cui]:
                    sim_vectors.append(unitvec(self.cui2context_vectors[_cui][context_type]))
                    sim_vectors_counts.append(self.cui2count_train[_cui])
                    sim_vectors_type_ids.append(self.cui2type_ids.get(_cui, {'unk'}))
                    sim_vectors_cuis.append(_cui)

            sim_data['sim_vectors'] = np.array(sim_vectors)
            sim_data['sim_vectors_counts'] = np.array(sim_vectors_counts)
            sim_data['sim_vectors_type_ids'] = np.array(sim_vectors_type_ids)
            sim_data['sim_vectors_cuis'] = np.array(sim_vectors_cuis)

        # Select appropriate concepts
        type_id_inds = np.arange(0, len(sim_data['sim_vectors_type_ids']))
        if len(type_id_filter) > 0:
            type_id_inds = np.array([], dtype=np.int32)
            for type_id in type_id_filter:
                type_id_inds = np.union1d(np.array([ind for ind, type_ids in enumerate(sim_data['sim_vectors_type_ids']) if type_id in type_ids]),
                        type_id_inds)
        cnt_inds = np.arange(0, len(sim_data['sim_vectors_counts']))
        if min_cnt > 0:
            cnt_inds = np.where(sim_data['sim_vectors_counts'] >= min_cnt)[0]
        # Intersect cnt and type_id 
        inds = np.intersect1d(type_id_inds, cnt_inds)

        mtrx = sim_data['sim_vectors'][inds]
        cuis = sim_data['sim_vectors_cuis'][inds]

        sims = np.dot(mtrx, unitvec(self.cui2context_vectors[cui][context_type]))

        sims_srt = np.argsort(-1*sims)

        # Create the return dict
        res = {}
        print()
        for ind, _cui in enumerate(cuis[sims_srt[0:topn]]):
            res[_cui] = {'name': self.cui2preferred_name.get(_cui, list(self.cui2names[_cui])[0]), 'sim': sims[sims_srt][ind],
                         'type_names': [self.addl_info['type_id2name'].get(cui, 'unk') for cui in self.cui2type_ids.get(_cui, ['unk'])],
                         'type_ids': self.cui2type_ids.get(_cui, 'unk'),
                         'cnt': self.cui2count_train[_cui]}

        return res



from __future__ import absolute_import, division, print_function

import pickle
import logging

import datasets


_CITATION = """\
@misc{kraljevic2020multidomain,
      title={Multi-domain Clinical Natural Language Processing with MedCAT: the Medical Concept Annotation Toolkit}, 
      author={Zeljko Kraljevic and Thomas Searle and Anthony Shek and Lukasz Roguski and Kawsar Noor and Daniel Bean and Aurelie Mascio and Leilei Zhu and Amos A Folarin and Angus Roberts and Rebecca Bendayan and Mark P Richardson and Robert Stewart and Anoop D Shah and Wai Keong Wong and Zina Ibrahim and James T Teo and Richard JB Dobson},
      year={2020},
      eprint={2010.01165},
      archivePrefix={arXiv},
      primaryClass={cs.CL}
}
"""

_DESCRIPTION = """\
Takes as input a pickled dict of pt2stream. The format should be:
    {'patient_id': (concept_cui, concept_count_for_patient, timestamp_of_first_occurrence_for_patient), ...}
"""

class PatientConceptStreamConfig(datasets.BuilderConfig):
    """ BuilderConfig for PatientConceptStream.

        Args:
            **kwargs: keyword arguments forwarded to super.
    """

    def __init__(self, **kwargs):
        super(PatientConceptStreamConfig, self).__init__(**kwargs)


class PatientConceptStream(datasets.GeneratorBasedBuilder):
    """PatientConceptStream: as input takes the patient to stream of concepts.

    TODO: Move the preparations scripts out of notebooks
    """

    BUILDER_CONFIGS = [
        PatientConceptStreamConfig(
            name="pickle",
            version=datasets.Version("1.0.0", ""),
            description="Pickled output from Temporal dataset preparation scripts",
        ),
    ]


    def _info(self):
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "patient_id": datasets.Value("string"),
                    "stream": [datasets.Value('string')],
                }
            ),
            # No default supervised_keys (as we have to pass both question
            # and context as input).
            supervised_keys=None,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                gen_kwargs={
                    "filepath": self.config.data_files,
                },
            ),
        ]

    def _generate_examples(self, filepath):
        """This function returns the examples in the raw (text) form."""
        logging.info("generating examples from = %s", filepath)
        with open(filepath, 'rb') as f:
            pt2stream = pickle.load(f)
            for pt, stream in pt2stream.items():
                out_stream = []
                year = -1
                for data in stream:
                    # 0 - CUI, 1 - CNT, 2 - TIME, 3 - Pt age in Years
                    if data[3] > year:
                        out_stream.append(str(data[3]))
                        year = data[3]

                    out_stream.append(data[0])

                yield pt, {'patient_id': str(pt),
                           'stream': out_stream}



import pandas as pd
import regex
import logging


def normalize_date(date, id, start, end):
    """ Normalizes different dates encountered in the clinical notes.
    Current accepted formats:
        28 Feb 2913 04:50
        Thu 28 Feb 2013 04:50
        28-Feb-2013 04:50
    Output:
        28 Feb 2013 04:50
    """

    if '-' in date:
        date = date.replace("-", " ").strip()
    elif date.strip()[0].isalpha():
        date = date[date.index(' '):].strip()
    elif date.strip()[0].isnumeric():
        # all good
        date = date.strip()
    else:
        logging.warning("Unsuported date format: {} for id: {} with start: {}, end: {}".format(date, id, start, end))
        return None

    return date


def split_one_note(id, text):
    """ Splits the text of one note by date.

    Return:
        split_note (List[Dict]):
            Returns a list of dictionary in the format: {'start': <start char of the specific note in the big one>,
                                                         'end': <end char of the specifc note in the big one>,
                                                         'text': <text of the specific note>,
                                                         'date': <date of the specific note>}
    """
    r = r'\n\w{0,5}\s*\d{1,2}(\s|-)[a-zA-Z]{3,5}(\s|-)\d{4}\s+\d{2}\:\d{2}'
    dates = regex.finditer(r, text)
    start = 0
    end = -1
    split_note = []
    previous_date = None

    for date in dates:
        if start == 0:
            start = date.span()[0]
            previous_date = date.captures()[0]
        elif previous_date is None or date.captures()[0] != previous_date:
            end = date.span()[0]
            note_text = text[start:end]
            if 'entered on -' in note_text.lower():
                if len(regex.findall(r'entered on -', note_text)) > 1:
                    logging.warning("Possible problems for span with start: {} and end: {} for note with id: {}".format(start, end, id))
                split_note.append({'start': start, 'end': end, 'text': note_text, 'date': normalize_date(previous_date, id, start, end)})
                start = end
                previous_date = date.captures()[0]
    # Add the last note
    if previous_date is not None and 'entered on -' in text[start:].lower():
        split_note.append({'start': start, 'end': len(text), 'text': text[start:], 'date': normalize_date(previous_date, id, start, len(text))})
    else:
        logging.warning("No date/entered-on detected for id: {} wth start: {}, end: {} and text:\n{}...".format(id, start, end, text[0:300]))

    return split_note


def split_clinical_notes(clinical_notes):
    """ Splits clinical notes.

    Args:
        clinical_notes(dict):
            Dictionary in the form {<clinical_note_id>: <text>, ...}
    """
    split_notes = {}
    for id, text in clinical_notes.items():
        split_notes[id] = split_one_note(id, text)
    return split_notes



#import hunspell
import re
from collections import Counter
from spacy.tokens import Span
import spacy
import os

CONTAINS_NUMBER = re.compile('[0-9]+')


class BasicSpellChecker(object):
    r'''
    '''
    def __init__(self, cdb_vocab, config, data_vocab=None):
        self.vocab = cdb_vocab
        self.config = config
        self.data_vocab = data_vocab


    def P(self, word):
        "Probability of `word`."
	# use inverse of rank as proxy
	# returns 0 if the word isn't in the dictionary
        cnt = self.vocab.get(word, 0)
        if cnt != 0:
            return -1 / cnt
        else:
            return 0


    def __contains__(self, word):
        if word in self.vocab:
            return True
        elif self.data_vocab is not None and word in self.data_vocab:
            return False
        else:
            return False


    def fix(self, word):
        "Most probable spelling correction for word."
        fix = max(self.candidates(word), key=self.P)
        if fix != word:
            return fix
        else:
            return None

    def candidates(self, word):
        "Generate possible spelling corrections for word."
        if self.config.general['spell_check_deep']:
            # This will check a two letter edit distance
            return (self.known([word]) or self.known(self.edits1(word)) or self.known(self.edits2(word)) or [word])
        else:
            # Will check only one letter edit distance
            return (self.known([word]) or self.known(self.edits1(word))  or [word])


    def known(self, words):
        "The subset of `words` that appear in the dictionary of WORDS."
        return set(w for w in words if w in self.vocab)


    def edits1(self, word):
        "All edits that are one edit away from `word`."
        letters    = 'abcdefghijklmnopqrstuvwxyz'
        splits     = [(word[:i], word[i:])    for i in range(len(word) + 1)]
        deletes    = [L + R[1:]               for L, R in splits if R]
        transposes = [L + R[1] + R[0] + R[2:] for L, R in splits if len(R)>1]
        replaces   = [L + c + R[1:]           for L, R in splits if R for c in letters]
        inserts    = [L + c + R               for L, R in splits for c in letters]
        return set(deletes + transposes + replaces + inserts)


    def edits2(self, word):
        "All edits that are two edits away from `word`."
        return (e2 for e1 in self.edits1(word) for e2 in self.edits1(e1))


    def edits3(self, word):
        "All edits that are two edits away from `word`."
        # Do d3 edits
        pass


class TokenNormalizer(object):
    r''' Will normalize all tokens in a spacy document.

    Args:
        config
        spell_checker
    '''

    def __init__(self, config, spell_checker=None):
        self.config = config
        self.spell_checker = spell_checker
        self.nlp = spacy.load(config.general['spacy_model'], disable=config.general['spacy_disabled_components'])

    def __call__(self, doc):
        for token in doc:
            if len(token.lower_) < self.config.preprocessing['min_len_normalize']:
                token._.norm = token.lower_
            elif (self.config.preprocessing.get('do_not_normalize', set())) and token.tag_ is not None and \
                     token.tag_ in self.config.preprocessing.get('do_not_normalize'):
                token._.norm = token.lower_
            elif token.lemma_ == '-PRON-':
                token._.norm = token.lemma_
                token._.to_skip = True
            else:
                token._.norm = token.lemma_.lower()

            if self.config.general['spell_check']:
                # Fix the token if necessary
                if len(token.text) >= self.config.general['spell_check_len_limit'] and not token._.is_punct \
                        and token.lower_ not in self.spell_checker and not CONTAINS_NUMBER.search(token.lower_):
                    fix = self.spell_checker.fix(token.lower_)
                    if fix is not None:
                        tmp = self.nlp(fix)[0]
                        if len(token.lower_) < self.config.preprocessing['min_len_normalize']:
                            token._.norm = tmp.lower_
                        else:
                            token._.norm = tmp.lemma_.lower()
        return doc



from spacy.tokens import Token, Doc, Span
from medcat.utils.normalizers import TokenNormalizer
import spacy
import os

class Pipe(object):
    r''' A wrapper around the standard spacy pipeline.

    Args:
        tokenizer (`spacy.tokenizer.Tokenizer`):
            What will be used to split text into tokens, can be anything built as a spacy tokenizer.
        config (`medcat.config.Config`):
            Global config for medcat.

    Properties:
        nlp (spacy.language.<lng>):
            The base spacy NLP pipeline.
    '''
    def __init__(self, tokenizer, config):
        self.nlp = spacy.load(config.general['spacy_model'], disable=config.general['spacy_disabled_components'])
        if config.preprocessing['stopwords'] is not None:
            self.nlp.Defaults.stop_words = set(config.preprocessing['stopwords'])
        self.nlp.tokenizer = tokenizer(self.nlp)


    def add_tagger(self, tagger, name, additional_fields=[]):
        r''' Add any kind of a tagger for tokens.

        Args:
            tagger (`object/function`):
                Any object/function that takes a spacy doc as an input, does something
                and returns the same doc.
            name (`str`):
                Name for this component in the pipeline.
            additional_fields (`List[str]`):
                Fields to be added to the `_` properties of a token.
        '''
        self.nlp.add_pipe(tagger, name='tag_' + name, first=True)
        # Add custom fields needed for this usecase
        Token.set_extension('to_skip', default=False, force=True)

        # Add any additional fields that are required
        for field in additional_fields:
            Token.set_extension(field, default=False, force=True)


    def add_token_normalizer(self, config, spell_checker=None):
        token_normalizer = TokenNormalizer(spell_checker=spell_checker, config=config)
        self.nlp.add_pipe(token_normalizer, name='token_normalizer', last=True)

        # Add custom fields needed for this usecase
        Token.set_extension('norm', default=None, force=True)


    def add_ner(self, ner):
        r''' Add NER from CAT to the pipeline, will also add the necessary fields
        to the document and Span objects.

        '''
        self.nlp.add_pipe(ner, name='cat_ner', last=True)

        Doc.set_extension('ents', default=[], force=True)
        Span.set_extension('confidence', default=-1, force=True)
        Span.set_extension('id', default=0, force=True)
        # Do not set this property if a vocabulary apporach is not used, this name must
        #refer to a name2cuis in the cdb.
        Span.set_extension('detected_name', default=None, force=True)
        Span.set_extension('link_candidates', default=None, force=True)


    def add_linker(self, linker):
        r''' Add entity linker to the pipeline, will also add the necessary fields
        to Span object.

        linker (object/function):
            Any object/function created based on the requirements for a spaCy pipeline components. Have
            a look at https://spacy.io/usage/processing-pipelines#custom-components
        '''
        self.nlp.add_pipe(linker, name='cat_linker', last=True)
        Span.set_extension('cui', default=-1, force=True)
        Span.set_extension('context_similarity', default=-1, force=True)


    def add_meta_cat(self, meta_cat, name):
        self.nlp.add_pipe(meta_cat, name=name, last=True)

        # Only the meta_anns field is needed, it will be a dictionary 
        #of {category_name: value, ...}
        Span.set_extension('meta_anns', default=None, force=True)


    def __call__(self, text):
        return self.nlp(text)



def encode_examples(left_context, center_context, right_context, tokenizer, max_seq_len):
    left_encoded = tokenizer(left_context)
    center_encoded = tokenizer(center_context)
    right_encoded = tokenizer(right_context)

    input_ids = left_encoded['input_ids']
    attention_mask = left_encoded['attention_mask']
    center_positions = [min(len(x), max_seq_len) for x in input_ids]

    for encoding in [center_encoded, right_encoded]:
        for i in range(len(input_ids)): input_ids[i].extend(encoding['input_ids'][i])
        for i in range(len(input_ids)): attention_mask[i].extend(encoding['attention_mask'][i])

    return {'input_ids': input_ids, 'center_positions': center_positions, 'attention_mask': attention_mask}
